from typing import Optional

from hoss.error import *
from hoss.namespace import Namespace
from hoss.core import CoreService
from hoss.auth import AuthService

import urllib.parse


def connect(server_url: str, auth_instance: Optional[AuthService] = None) -> CoreService:
    """Connect to a server

    Args:
        server_url: URL to the server, including the protocol (e.g. https://hoss.myserver.com)
        auth_instance: Optionally provide an already configured auth instance. This can be useful when
                       interacting with multiple Hoss servers that each run their own auth service.

    Returns:
        A CoreService instance, which is the primary interface to a server.
    """
    return CoreService(server_url, auth_instance)


def resolve(uri, auth_instance: Optional[AuthService] = None):
    """Resolve a Hoss URI into a DatasetRef. This lets you directly load any ref.

    The Hoss URI format is `hoss+<server>:<namespace>:<dataset>/<object key>` and can be retrieved
    from any DatasetRef instance by the `uri` property.

    An example URI would be `hoss+https://hoss.myserver.com:default:example-ds/my-file.bin`

    Args:
        uri: Hoss URI formatted string
        auth_instance: Optionally provide an already configured auth instance. This can be useful when
                       interacting with multiple Hoss servers that each run their own auth service.

    Returns:
        A populated DatasetRef if the URI is valid
    """
    uri = urllib.parse.urlparse(uri)

    if not uri.scheme.lower().startswith("hoss"):
        raise ValueError("URI is not a valid Hoss URI")

    try:
        _, protocol = uri.scheme.lower().split("+")
        host, namespace_name, dataset_name = uri.netloc.split(":")
    except:
        raise ValueError("URI is not a valid Hoss URI")

    s = CoreService(f"{protocol}://{host}", auth_instance=auth_instance)
    ns = s.get_namespace(namespace_name)

    return ns.get_dataset(dataset_name) / uri.path



class HossException(Exception):
    """A general exception raised by the hoss package"""
    pass


class NotAuthorizedException(HossException):
    """You are not authorized to access this API/resource"""
    pass


class NotFoundException(HossException):
    """The specified resource was not found"""
    pass


class AlreadyExistsException(HossException):
    """The specified resource already exists"""
    pass


class ServerCheckError(Exception):
    """An error occurred while trying to fetch the server version"""
    pass



from typing import Tuple
import requests

from hoss.error import ServerCheckError

# Set the version of the library here
__version__ = "0.6.6"


def get_server_version(server_url: str) -> Tuple[str, str]:
    """Function to fetch the discover endpoint from the local server

    Returns:
        a dict with the version and build hash
    """
    try:
        resp = requests.get(f"{server_url}/core/v1/discover")
        if resp.status_code != 200:
            raise ServerCheckError(f"Failed to load version information from server. Status Code: {resp.status_code}")
    except requests.exceptions.ConnectionError:
        raise ServerCheckError(f"Failed to load version information from server. Could not connect to server.")

    data = resp.json()

    return data['version'], data['build']



import click
from hoss.tools.upload import upload_directory


@click.command()
@click.argument('dataset_name', metavar='DATASET_NAME', type=str, nargs=1)
@click.argument('directory', type=str, nargs=1)
@click.option('--namespace', '-n', type=str, default="default", show_default=True,
              help="Namespace that contains the dataset")
@click.option('--endpoint', '-e', type=str, default="http://localhost", show_default=True,
              help="Hoss server root endpoint")
@click.option('--skip', '-s', type=str, default="",
              help="Optional regular expression used to filter out files to skip (e.g. myprefix.*\\.txt)")
@click.option('--max_concurrency', '-c', type=int, default=10, show_default=True,
              help="Maximum number of concurrent s3 API transfer operations, defaults to 10.")
@click.option('--multipart_threshold', type=int, default=32, show_default=True,
              help="Threshold in megabytes for which transfers will be split into multiple parts, defaults to 32MB")
@click.option('--multipart_chunk_size', type=int, default=32, show_default=True,
              help="Size in megabytes for each multipart chunk, if used. Defaults to 32MB")
@click.option('--metadata', '-m', type=str, multiple=True, default=list(),
              help="Object metadata key-value pair(s) applied to every object uploaded."
                   " You may specify multiple values by repeating the option (e.g. -m foo=bar -m fizz=buzz")
@click.pass_context
def upload(ctx, dataset_name: str, directory: str, namespace: str,
           endpoint: str, skip: str, max_concurrency: int, multipart_threshold: int, multipart_chunk_size: int,
           metadata: str):
    """Upload files in a directory to an existing dataset"""
    # Convert metadata to dict if it is set, otherwise set to None
    if metadata:
        metadata_pairs = [item.split("=") for item in metadata]
        metadata_dict = dict((k.lower(), v.lower()) for k, v in metadata_pairs)
    else:
        metadata_dict = None

    upload_directory(dataset_name, directory, namespace, endpoint, skip,
                     max_concurrency, multipart_threshold, multipart_chunk_size, metadata_dict)



import click

from hoss.version import __version__, get_server_version
from hoss.error import ServerCheckError
from hoss.console import console


@click.command()
@click.option('--endpoint', '-e', type=str, default="http://localhost", show_default=True,
              help="Hoss server endpoint")
def version(endpoint):
    """Print client library and server version info

    /f

    Returns:
        None
    """
    console.print(f"\nHoss client library: v{__version__}")

    try:
        server_version, build = get_server_version(endpoint)
        console.print(f"Hoss server: v{server_version} (build {build})\n")
    except ServerCheckError as err:
        console.print(f"Hoss server: NOT AVAILABLE\n")
        console.print(f"Failed to fetch server version from {endpoint}. If the server is located at a different "
                      f"location, provide the URL via the '--endpoint' option "
                      f"(e.g. '--endpoint https://hos.mydomain.com')")
        console.print(f" - Error: {err}\n\n")




from typing import Optional
from urllib.parse import urlparse

import requests
from packaging import version

from hoss.version import __version__
from hoss.error import *
from hoss.auth import AuthService


# Currently must be a 0.2.1 or later server
MIN_SUPPORTED_SERVER_VERSION = "0.2.1"
MAX_SUPPORTED_SERVER_VERSION = "0.2.99999999"


class CoreAPI(object):
    """Class to interact the Hoss APIs. This class contains private methods that are used by other classes
    to interact with Hoss server APIs"""
    def __init__(self, server_url: str, auth_instance: Optional[AuthService] = None):
        """A root class to interact with the Hoss core API
        
        Args:
            server_url: root URL to the server, including the protocol
            auth_instance: an optional AuthService instance that will be used instead of using the server_url to
                           instantiate one. This is useful if working with 2 linked servers without having to switch
                           HOS_PAT env vars.
        """
        parts = urlparse(server_url)
        self.host = parts.netloc
        self.base_url = f"{parts.scheme}://{parts.netloc}"

        self._verify_server_version()

        if auth_instance:
            self.auth = auth_instance
        else:
            # Discover the auth endpoint you should be using by calling the core service discovery endpoint
            response = requests.get(f"{self.base_url}/core/v1/discover")
            if response.status_code != 200:
                raise HossException(f"Failed to reach discover endpoint for server '{self.base_url}' when trying "
                                    f"to initialize auth")
            data = response.json()
            self.auth = AuthService(data['auth_service'])

    def _request(self, method: str, path, json=None) -> Optional[dict]:
        """Method to make a request to the API

        Args:
            method: HTTP method
            path: relative path in the API (i.e. after <host>/core/v1). Includes the leading slash
            json: dictionary to encode as json in the POST body

        Returns:
            dictionary of the json formatted response, if present
        """
        kwargs = {
            "headers": {
                "Access-Control-Request-Method": method,
                "Origin": self.base_url,
                **self.auth.headers(),
            },
        }
        if json:
            kwargs["json"] = json

        resp = requests.request(method, f"{self.base_url}/core/v1{path}", **kwargs)
        if resp.ok:
            if resp.text == "":
                return None
            else:
                return resp.json()
        if resp.status_code == 401:
            raise NotAuthorizedException()
        if resp.status_code == 404:
            raise NotFoundException()
        if resp.status_code == 403:
            raise AlreadyExistsException()
        
        try:
            error = resp.json().get("error")

            if error == "Resource not found":
                raise NotFoundException()
            elif error == "Resource already exists":
                raise AlreadyExistsException()
            else:
                raise HossException(error)
        except HossException:
            raise  # don't modify the exceptions just raised
        except:
            raise HossException(f"{resp.status_code} {resp.reason}: {resp.text}")

    def _sync_user_groups(self) -> None:
        """Helper method to synchronize a user's groups. Useful if group membership changed and you need to
        update your groups in the JWT"""
        self._request("PUT", "/user/sync")

    def _verify_server_version(self) -> None:
        """Helper method to verify that the version of this library is compatible with the server version

        If the server version is not compatible, a `HossException` will be raised.

        Returns:
            None
        """
        kwargs = {
            "headers": {
                "Access-Control-Request-Method": "GET",
                "Origin": self.base_url
            },
        }
        response = requests.request("GET", f"{self.base_url}/core/v1/discover", **kwargs)
        if response.status_code != 200:
            raise HossException(f"Failed to fetch server version. Status Code: {response.status_code}")

        data = response.json()
        server_version = version.parse(data['version'])

        if server_version < version.parse(MIN_SUPPORTED_SERVER_VERSION):
            raise HossException(f"The current client version ({__version__}) is not compatible with this server."
                               f" (Server version: {data['version']}, Minimum supported server version:"
                               f" {MIN_SUPPORTED_SERVER_VERSION}. Downgrade the client and try again.")

        if server_version > version.parse(MAX_SUPPORTED_SERVER_VERSION):
            raise HossException(f"The current client version ({__version__}) is not compatible with this server."
                               f" (Server version: {data['version']}, Maximum supported server version:"
                               f" {MAX_SUPPORTED_SERVER_VERSION}. Upgrade the client and try again.")



from typing import Optional, TYPE_CHECKING, Dict, List

import io
import re
import fnmatch
import contextlib
import datetime

from botocore.exceptions import ClientError

from hoss.error import HossException
if TYPE_CHECKING:
    from hoss.namespace import Namespace


class DatasetRef:
    """A reference to an object in a dataset

    Attributes:
        namespace(Namespace): The namespace containing the dataset to which this this reference belongs
        parent(DatasetRef): the parent ref. For the Dataset itself, this will be None. For a file in a "folder", it will
                            be the folder.
        key(str): The key to the object in the dataset
        name(str): The name of the ref (i.e. the name of the object without the "path")
        uri(str): The Hoss URI to this ref
    """
    def __init__(self, namespace: 'Namespace', parent: Optional['DatasetRef'], dataset_name: str,
                 key: str, name: str,
                 etag: Optional[str], last_modified: Optional[datetime.datetime], size_bytes: Optional[int]):
        """Constructor

        Args:
            namespace: the Namespace object that contains this dataset
            parent: the parent ref. For the Dataset itself, this will be None.
            dataset_name: the name of the dataset that contains this ref
            key: the key to the ref in the object store. Note, this will be prefixed with the dataset name.
            name: the name of the ref (i.e. the name of the object without the "path")
            etag: the etag of the object
            last_modified: the last modified datetime of the object in the store
            size_bytes: the size of the object in bytes
        """
        if name == f".dataset.yaml":
            # We don't want anyone interacting with the internal .dataset.yaml Hoss file.
            # While this file only exists in the root of the dataset and we are effectively blocking
            # the filename anywhere in a dataset, this simplifies both this check and the list operation
            # which filters out this file. If future justification is strong enough, we could
            # block only this file in the root and then update the individual methods that list files
            # to also filter out the file, but it seems unlikely that anyone needs a `.dataset.yaml` file.
            raise HossException(".dataset.yaml is a protected Hoss internal filename.")

        self.namespace = namespace
        self.dataset_name = dataset_name
        self.parent = parent
        self.key = key
        self.name = name
        self.uri = f"hoss+{namespace.base_url}:{self.namespace.name}:{key}"

        self._etag = etag
        self._last_modified = last_modified
        self._size_bytes = size_bytes

        self._metadata = None

        if '/' in name[:-1]:
            base_key = key[:len(key)-len(name)]

            parts = name.split('/')
            part = parts.pop()
            if part == '':  # it's a directory
                self.name = parts.pop() + '/'
            else:
                self.name = part

            for part in parts:
                base_key = f"{base_key}{part}/"
                self.parent = DatasetRef(self.namespace, self.parent, self.dataset_name, base_key, part + '/',
                                         etag=None, last_modified=None, size_bytes=None)

    def _populate_head_data(self) -> None:
        """Helper method to update the instance with data from a HEAD request"""
        try:
            data = self.namespace.object_store.head_object(self.namespace.name, self.namespace.bucket, self.key)
        except ClientError as err:
            if err.response['Error']['Code'] == '404':
                # If the object isn't found...it doesn't exist! Just return since no metadata exists.
                return
            else:
                raise

        self._metadata = data['metadata']

        # Metadata is case-insensitive, so we always enforce this for the user.
        if self._metadata:
            self._metadata = dict((k.lower(), str(v).lower()) for k, v in self._metadata.items())
        self._etag = data['etag']
        self._size_bytes = data['size_bytes']
        self._last_modified = data['last_modified']

    @property
    def metadata(self) -> dict:
        """Getter for the ref's metadata key-value pairs"""
        if not self._metadata:
            self._populate_head_data()
        return self._metadata

    @property
    def etag(self) -> str:
        """Getter for the etag of the object. The etag is a hash and will change when the file's contents change"""
        if not self._etag:
            self._populate_head_data()
        return self._etag

    @property
    def last_modified(self) -> datetime.datetime:
        """Getter for the last modified datetime in the object store backend"""
        if not self._last_modified:
            self._populate_head_data()

        return self._last_modified

    @property
    def size_bytes(self) -> int:
        """Getter for the size of the objects in bytes"""
        if not self._last_modified:
            self._populate_head_data()

        return self._size_bytes

    def __repr__(self):
        return f"<DatasetRef: {self.key}>"

    def __truediv__(self, name) -> 'DatasetRef':
        """Method to implement the `/` operator like how pathlib lets you build paths"""
        if name[0] == '/':
            name = name[1:]

        if self.is_dir():
            key = self.key + name
        else:
            key = self.key + '/' + name

        return DatasetRef(self.namespace, self, self.dataset_name, key, name,
                          etag=None, last_modified=None, size_bytes=None)

    def copy_to(self, target) -> None:
        """Method to copy the contents of 1 reference to another

        Args:
            target(DatasetRef): A target ref to copy to

        Returns:
            None
        """
        if type(target) != DatasetRef:
            raise ValueError("target should be a DatasetRef")

        self.namespace.object_store.copy_object(self.namespace.name, self.namespace.bucket, self.key, target.key)

    def exists(self) -> bool:
        """Check if an ref exists in the object store

        Returns:
            true if the object exists, false if it does not
        """
        return self.namespace.object_store.object_exists(self.namespace.name, self.namespace.bucket, self.key)

    def glob(self, pattern) -> 'DatasetRef':
        """Method to apply unix style pathname pattern expansion while iterating through a dataset

        If the pattern starts with `**/` then `rglob` will be used internally

        Args:
            pattern: The pattern to provide

        Yields:
            DatasetRef instance
        """
        if pattern.startswith("**/"):
            # cannot just return the rglob generator as this method is also a generator
            for ref in self.rglob(pattern):
                yield ref
            return

        compiled = re.compile(fnmatch.translate(pattern))
        for ref in self.namespace.object_store.list_objects(self.namespace.name, self.namespace.bucket, self.key):
            if compiled.match(ref["name"]) is not None:
                yield DatasetRef(self.namespace, self, self.dataset_name, ref['key'], ref['name'],
                                 ref['etag'], ref['last_modified'], ref['size_bytes'])

    def is_file(self) -> bool:
        """Method to check if a reference is a file

        This is done assuming the convention that "directories" end in `/` and are virtually represented client side.

        Returns:
            True if this ref is a file, False if it is a directory
        """
        return not self.is_dir()

    def is_dir(self) -> bool:
        """Method to check if a reference is a directory

        This is done assuming the convention that "directories" end in `/` and are virtually represented client side.
        To create a "directory" you must write an object into the directory.

        Returns:
            True if this ref is a file, False if it is a directory
        """
        return self.key[-1] == '/'

    def iterdir(self) -> List['DatasetRef']:
        """Method to iterate through references at the "level" represented by this DatasetRef

        If this is the dataset itself, this method will iterate through the root of the dataset. If this is some
        other "directory" in the dataset, this method will iterate through the directory.

        Returns:
            List of DatasetRefs in this "directory"
        """
        return [DatasetRef(self.namespace, self, self.dataset_name, ref['key'], ref['name'],
                           ref['etag'], ref['last_modified'], ref['size_bytes'])
                for ref in self.namespace.object_store.list_objects(self.namespace.name,
                                                                    self.namespace.bucket,
                                                                    self.key)]

    def move(self, target: 'DatasetRef') -> None:
        """Method to move this reference to another

        Args:
            target: The destination DatasetRef that this DatasetRef should be moved to

        Returns:
            None
        """
        if type(target) != DatasetRef:
            raise ValueError("target should be a DatasetRef")

        self.copy_to(target)
        try:
            self.remove()
        except Exception as err:
            err_msg = f"Failed to remove source object during move operation: {err}"
            try:
                target.remove()
            except Exception as err:
                err_msg = f"{err_msg}. Failed to remove target object while rolling back move operation: {err}"
            finally:
                raise Exception(err_msg)

    @contextlib.contextmanager
    def open(self, mode: str = "r", buffering=-1, encoding=None, errors=None, newline=None):
        """Open a file-like object in a context manager that you can read/write to.

        Args:
            mode:
            buffering:
            encoding:
            errors:
            newline:

        Returns:

        """
        buf = io.BytesIO()

        if 'r' in mode or 'a' in mode:
            self.read_to(buf)

        if 'r' in mode:
            buf.seek(0, 0)  # seek beginning

        if 'a' in mode:
            buf.seek(0, 2)  # seek end

        if 'b' not in mode:
            wrap = io.TextIOWrapper(buf,
                                    encoding=encoding,
                                    errors=errors,
                                    newline=newline,
                                    line_buffering=(buffering == 1),
                                    write_through=(buffering == 0))

        try:
            if 'b' not in mode:
                yield wrap
            else:
                yield buf
        finally:
            if 'b' not in mode:
                wrap.flush()

            if 'r' not in mode and '+' not in mode:
                buf.seek(0, 0) # rewind so all data is uploaded
                self.write_from(buf)

            if not buf.closed:
                buf.close()

    def read_bytes(self) -> bytes:
        """Read the contents of this ref in bytes

        Returns:
            bytes
        """
        return self.namespace.object_store.get_object(self.namespace.name, self.namespace.bucket, self.key)

    def read_text(self, encoding='utf-8') -> str:
        """Read the contents of this object as a string

        Args:
            encoding: The encoding to use when decoding the binary object data

        Returns:
            a string containing the object's contents
        """
        return str(self.read_bytes(), encoding=encoding)

    def read_to(self, fh_or_name):
        """Method to read to a file handle to a file-like object or a file specified by an absolute path

        Args:
            fh_or_name: A file handle to a file-like object in binary mode or an absolute path string

        Returns:

        """
        if type(fh_or_name) == str:
            return self.namespace.object_store.download_file(self.namespace.name, self.namespace.bucket,
                                                             self.key, fh_or_name)
        else:
            if hasattr(fh_or_name, "mode") and 'b' not in fh_or_name.mode:
                raise ValueError("Can only read_to as binary data")
            if type(fh_or_name) == io.StringIO:
                raise ValueError("Can only read_to as binary data")
            return self.namespace.object_store.download_fileobj(self.namespace.name, self.namespace.bucket,
                                                                self.key, fh_or_name)

    def remove(self) -> None:
        """Remove a reference, ignoring if the reference doesn't exist

        Returns:
            None
        """
        self.unlink(missing_ok=True)

    def rglob(self, pattern):
        """Method to recursively apply unix style pathname pattern expansion while iterating through a dataset

        Args:
            pattern: pattern to match

        Yields:
            DatasetRef
        """
        if pattern.startswith("**/"):
            pattern = pattern[1:]

        compiled = re.compile(fnmatch.translate(pattern))
        for ref in self.namespace.object_store.list_objects(self.namespace.name,
                                                            self.namespace.bucket,
                                                            self.key, recursive=True):
            key_part = ref['key'][len(self.key)-1:]  # leave the initial '/'
            if compiled.match(key_part) is not None:
                yield DatasetRef(self.namespace, self, self.dataset_name, ref['key'], ref['name'],
                                 ref['etag'], ref['last_modified'], ref['size_bytes'])

    def touch(self, exists_ok=True) -> None:
        """Write an empty object if the object does not exist

        Touch only has en effect if the object does not exist.

        Args:
            exists_ok: If False and the object already exists, an exception is raised

        Returns:
            None
        """
        if self.exists():
            if not exists_ok:
                raise FileExistsError(f"File {self.uri} already exists")
        else:
            self.write_bytes(b'')

    def unlink(self, missing_ok=False) -> None:
        """Method to remove an object

        Args:
            missing_ok: If False (default), an exception will be raised if the reference does not exist

        Returns:
            None
        """
        if self.exists():
            self.namespace.object_store.delete_object(self.namespace.name, self.namespace.bucket, self.key)
        else:
            if not missing_ok:
                raise FileNotFoundError(f"File {self.uri} doesn't exist")

    def write_bytes(self, data, metadata: Optional[dict] = None):
        """Method to write bytes to a ref

        Args:
            data: binary data to write to the ref
            metadata: dictionary of metadata tags of type Dict[str, str]. If None and the reference already has
                      metadata tags, they will be persisted. If set to an empty dict() all keys will be removed.

        Returns:
            None
        """
        if metadata is None and self.metadata is not None:
            # If the user hasn't provided metadata, but this reference already has some, don't lose it
            metadata = self.metadata
        elif metadata is not None:
            # If metadata is set, update the local reference to save a HEAD operation
            # Also lower case keys and values since metadata is case insensitive.
            metadata = dict((k.lower(), str(v).lower()) for k, v in metadata.items())
            self._metadata = metadata

        # Reset properties we'd want to re-fetch if the user requests them after the write
        self._etag = None
        self._last_modified = None
        self._size_bytes = None

        self.namespace.object_store.put_object(self.namespace.name, self.namespace.bucket, self.key, data,
                                               metadata=metadata)

    def write_text(self, data: str, encoding: str = 'utf-8', metadata: Optional[Dict[str, str]] = None) -> None:
        """Method to write text to a ref

        Args:
            data: string to write to the ref
            encoding: encoding to use when converting from text to binary
            metadata: dictionary of metadata tags of type Dict[str, str]. If None and the reference already has
                      metadata tags, they will be persisted. If set to an empty dict() all keys will be removed.

        Returns:
            None
        """
        self.write_bytes(data.encode(encoding=encoding), metadata=metadata)

    def write_from(self, fh_or_name, metadata: Optional[dict] = None):
        """Method to write from a file handle OR a string containing an absolute file path

        Args:
            fh_or_name: file handle or absolute path to a file
            metadata: dictionary of metadata tags of type Dict[str, str]. If None and the reference already has
                      metadata tags, they will be persisted. If set to an empty dict() all keys will be removed.

        Returns:
            None
        """
        if metadata is None and self.metadata is not None:
            # If the user hasn't provided metadata, but this reference already has some, don't lose it
            metadata = self.metadata
        elif metadata is not None:
            # If metadata is set, update the local reference to save a HEAD operation
            # Also lower case keys and values since metadata is case insensitive.
            metadata = dict((k.lower(), str(v).lower()) for k, v in metadata.items())
            self._metadata = metadata

        # Reset properties we'd want to re-fetch if the user requests them after the write
        self._etag = None
        self._last_modified = None
        self._size_bytes = None

        if type(fh_or_name) == str:
            return self.namespace.object_store.upload_file(self.namespace.name, self.namespace.bucket,
                                                           self.key, fh_or_name, metadata=metadata)
        else:
            if hasattr(fh_or_name, "mode") and 'b' not in fh_or_name.mode:
                raise ValueError("Can only write_from as binary data")
            if type(fh_or_name) == io.StringIO:
                raise ValueError("Can only write_from as binary data")
            return self.namespace.object_store.upload_fileobj(self.namespace.name, self.namespace.bucket,
                                                              self.key, fh_or_name,  metadata=metadata)



from typing import Optional, List
import os
import time
from dataclasses import dataclass
from enum import Enum

import requests
import jwt

from hoss.error import *


class Role(Enum):
    """Enum for user roles"""
    USER = 0
    PRIVILEGED = 1
    ADMIN = 2


@dataclass
class User:
    """Dataclass for representing users"""
    username: str
    email: str
    role: Role


@dataclass
class Group:
    """Dataclass for representing groups"""
    name: str
    description: str
    members: Optional[List[User]]


class AuthService(object):
    """A class to represent a Hoss server's Auth Service"""
    def __init__(self, auth_service_endpoint: str) -> None:
        """Constructor

        Args:
            auth_service_endpoint: Endpoint to the desired auth service (e.g. http://localhost/auth/v1)
        """
        self.endpoint = auth_service_endpoint
        self.pat = os.environ.get("HOSS_PAT")
        self.jwt: Optional[str] = None
        self.jwt_exp_seconds: Optional[int] = None
        self.jwt_refresh_at: Optional[int] = None

        # Get a JWT on load. Also useful because it quickly verifies server access
        self._get_jwt()

    def _get_jwt(self) -> None:
        """Method to exchange a PAT for a JWT or alternatively load it directly from an env var `HOSS_JWT`

        Returns:
            None
        """
        # Set jwt property
        if self.pat is None:
            self.jwt = os.environ.get("HOSS_JWT")
            if self.jwt is None:
                raise HossException("env var 'HOSS_PAT' or 'HOSS_JWT' must be set to authenticate with a server.")
        else:
            headers = {"Authorization": f"Bearer {self.pat}"}
            try:
                resp = requests.request("POST", f"{self.endpoint}/pat/exchange/jwt", headers=headers)
            except requests.exceptions.ConnectionError:
                raise HossException(f"Cannot reach Hoss auth service for server '{self.endpoint}'. "
                                    f"Verify your network connection and try again.")

            if not resp.ok:
                print(f"{resp.status_code} {resp.reason}: {resp.text}")
                raise HossException("Could not retrieve JWT using PAT: " + resp.text)

            self.jwt = resp.json()["id_token"]

        if not self.jwt:
            raise HossException("Failed to load JWT")

        claims = jwt.decode(self.jwt,
                            options={"verify_signature": False},
                            audience="hoss",
                            issuer="hoss auth")
        self.jwt_exp_seconds = claims["exp"] - claims["iat"]
        self.jwt_refresh_at = claims["iat"] + self.jwt_exp_seconds / 2
        if self._has_jwt_expired():
            raise Exception("HOSS_JWT has expired")

    def _has_jwt_expired(self) -> bool:
        """Check if the JWT currently set in this instance has expired, using 3 hour buffer to allow up to 3 hours
        expiration for temporary STS credentials

        Returns:
            bool
        """
        return time.time() > self.jwt_refresh_at

    def headers(self) -> dict:
        """Method to populate a dict with request headers, primarily setting the Authorization header

        Returns:
            request headers
        """
        if self._has_jwt_expired():
            self._get_jwt()

        return {"Authorization": f"Bearer {self.jwt}"}

    # groups API
    def get_user(self, username) -> User:
        """Method to get a user's properties

        Args:
            username(str): username of the desired user

        Returns:
            A populated User instance
        """
        resp = requests.request("GET", f"{self.endpoint}/user/{username}", headers=self.headers())
        if not resp.ok:
            if "not found" in resp.text:
                raise NotFoundException()

            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not get user info: " + resp.text)

        data = resp.json()
        return User(data['username'],
                    data['email'],
                    Role[data['role'].upper()])

    @staticmethod
    def _parse_memberships(memberships: list) -> List[User]:
        """Helper method to parse group membership data into a list of User instances

        Args:
            memberships: dictionary of user group memberships returned from the API

        Returns:
            A list of User instances representing members of a group
        """
        if memberships:
            return [User(m['user']['username'],
                         m['user']['email'],
                         Role[m['user']['role'].upper()]) for m in memberships]
        else:
            return list()

    def get_group(self, group_name: str) -> Group:
        """Method to get a group and its properties

        Args:
            group_name: name of the group to fetch

        Returns:
            A populated Group instance
        """
        resp = requests.request("GET", f"{self.endpoint}/group/{group_name}", headers=self.headers())
        if not resp.ok:
            if "not found" in resp.text:
                raise NotFoundException()

            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not get group info: " + resp.text)
        data = resp.json()
        return Group(data['group_name'], data['description'],
                     members=self._parse_memberships(data.get('memberships')))

    def create_group(self, group_name: str, description: str = "No description provided") -> Group:
        """Method to create a new, empty group

        Args:
            group_name: Name of the group
            description: Description of the group

        Returns:
            Group
        """
        data = {"name": group_name, "description": description}
        resp = requests.request("POST", f"{self.endpoint}/group", json=data, headers=self.headers())
        if not resp.ok:
            if "already exists" in resp.text:
                raise AlreadyExistsException()

            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not create group: " + resp.text)

        return self.get_group(group_name)

    def delete_group(self, group_name: str) -> None:
        """Method to delete a group

        Args:
            group_name: name of the group to delete

        Returns:
            None
        """
        resp = requests.request("DELETE", f"{self.endpoint}/group/{group_name}", headers=self.headers())
        if not resp.ok:
            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not delete group: " + resp.text)

    def add_user_to_group(self, group_name: str, username: str) -> None:
        """Method to add a user to an existing group.

        Args:
            group_name: Name of the group
            username: Username of the user to add

        Returns:
            None
        """
        resp = requests.request("PUT", f"{self.endpoint}/group/{group_name}/user/{username}",
                                headers=self.headers())
        if not resp.ok:
            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not add user to group: " + resp.text)

    def remove_user_from_group(self, group_name: str, username: str) -> None:
        """Method to remove a user from a group

        Args:
            group_name: Name of the group
            username: Username of the user to remove

        Returns:
            None
        """
        resp = requests.request("DELETE", f"{self.endpoint}/group/{group_name}/user/{username}", headers=self.headers())
        if not resp.ok:
            print(f"{resp.status_code} {resp.reason}: {resp.text}")
            raise HossException("Could not remove user from group: " + resp.text)



import os
import hashlib

from boto3.s3.transfer import MB
from hoss.ref import DatasetRef


def hash_file(path: str, chunk_size: int = 8 * MB, multipart_threshold: int = 8 * MB) -> str:
    """Utility to hash a file like how S3 etags are generated.

    If the file is smaller than or equal to the multipart upload chunk size, it is assumed a single
    chunk was used. In this case, the etag is simply the hexdigest of the md5 hash of the file.

    If the file is larger than the multipart upload chunk size, it is assumed multiple parts
    were used to write the file. In this case, the etag is the hexdigest of the md5 hash
    of each part's md5 hash concatenated, with `-<number of parts>` appended.

    Args:
        path: path to the file to hash
        chunk_size: size in bytes that would be used to chunk a file during a multipart upload (default is 8MB)
        multipart_threshold: size in bytes that will trigger multipart uploads to be used (default is 8MB)

    Returns:
        etag formatted hash
    """
    info = os.stat(path)

    if info.st_size <= multipart_threshold:
        # Single part upload
        return hashlib.md5(open(path, 'rb').read()).hexdigest()
    else:
        # Multipart upload
        chunk_hashes = list()
        with open(path, 'rb') as f:
            while True:
                chunk_bytes = f.read(chunk_size)
                if not chunk_bytes:
                    break

                chunk_hashes.append(hashlib.md5(chunk_bytes))

        # A multi-part upload was used. Need to create combined hash
        combined_hash = b""
        for c in chunk_hashes:
            combined_hash += c.digest()
        return f"{hashlib.md5(combined_hash).hexdigest()}-{len(chunk_hashes)}"


def etag_does_match(ref: DatasetRef, path: str) -> bool:
    """Utility to check if an etag matches the contents of a local file

    Args:
        ref: the DatasetRef you are comparing against (this reference's etag will be used)
        path: path to the file to hash and compare

    Returns:
        true if contents are the same, false if they are not
    """
    local_etag = hash_file(path,
                           ref.namespace.object_store.get_multipart_chunk_size(),
                           ref.namespace.object_store.get_multipart_threshold())
    if ref.etag == f'"{local_etag}"':
        return True
    else:
        return False



from typing import Optional, List
from urllib.parse import urlparse
import datetime

from hoss.error import *
from hoss.auth import AuthService
from hoss.dataset import Dataset
from hoss.api import CoreAPI
from hoss.objectstore import ObjectStore


class Namespace(CoreAPI):
    def __init__(self, server_url: str, name: str, bucket_name: str,
                 description: Optional[str] = None,
                 object_store: Optional[ObjectStore] = None,
                 auth_instance: Optional[AuthService] = None) -> None:
        """A class to interact with the Hoss core service
        
        Args:
            server_url: root URL to the server, including the protocol
            name: name of the namespace
            auth_instance: an optional AuthService instance that will be used instead of using the
                           server_url to instantiate one
        """
        super().__init__(server_url, auth_instance)
        self.name = name
        self.description = description
        self.object_store = object_store
        self.bucket = bucket_name

    def __repr__(self):
        return f"<Namespace: {self.name} - {self.description}>"

    def get_sync_configuration(self) -> dict:
        """Method to get the sync configuration for the namespace

        Returns:
            a list of dictionaries with the fields "target_core_service", "target_namespace", and "sync_type"
        """
        response = self._request("GET", f"/namespace/{self.name}/sync")
        result = {"sync_enabled": len(response) > 0, "sync_targets": list()}
        for t in response:
            result['sync_targets'].append({"target_core_service": t["target_core_service"],
                                           "target_namespace": t["target_namespace"],
                                           "sync_type": t["sync_type"]})

        return result

    def enable_sync_target(self, target_server_url: str, sync_type: str,
                           target_namespace: Optional[str] = None) -> None:
        """Enable a sync target, linking the namespaces together.

        Args:
            target_server_url: URL to the target server (e.g. https://hoss.mydomain.com)
            sync_type: The sync type ('simplex' or 'duplex')
            target_namespace: The namespace name to sync with

        Returns:
            None
        """
        parts = urlparse(target_server_url)
        target_core_service = f"{parts.scheme}://{parts.netloc}/core/v1"

        if not target_namespace:
            # If the target namespace is not provided, set it to the same name as the source namespace
            target_namespace = self.name

        if sync_type not in ["simplex", "duplex"]:
            raise HossException(f"'sync_type' must be either 'simplex' or 'duplex'")

        data = {"target_core_service": target_core_service,
                "target_namespace": target_namespace,
                "sync_type": sync_type}
        self._request("PUT", f"/namespace/{self.name}/sync", json=data)

    def disable_sync_target(self, target_server_url: str, target_namespace: str) -> None:
        """Remove a sync target, stopping all synchronization between the namespaces

        Args:
            target_server_url: URL to the target server (e.g. https://hoss.mydomain.com)
            target_namespace: The namespace name to remove

        Returns:
            None
        """
        parts = urlparse(target_server_url)
        target_core_service = f"{parts.scheme}://{parts.netloc}/core/v1"

        data = {"target_core_service": target_core_service,
                "target_namespace": target_namespace,
                "sync_type": "simplex"}
        self._request("DELETE", f"/namespace/{self.name}/sync", json=data)

    def _populate_dataset(self, api_response: dict) -> Dataset:
        perms_processsed = {
            perm["group"]["group_name"]: perm["permission"]
            for perm in api_response["permissions"]
        }

        return Dataset(self,
                       api_response["name"],
                       api_response["description"],
                       datetime.datetime.strptime(api_response['created'], "%Y-%m-%dT%H:%M:%S.%fZ"),
                       api_response['root_directory'],
                       api_response["owner"]["username"],
                       self.bucket,
                       perms_processsed)

    def list_datasets(self) -> List[Dataset]:
        """List Datasets within the namespace

        Returns:
            A list of Dataset instances
        """
        response = self._request("GET", f"/namespace/{self.name}/dataset")
        return [self._populate_dataset(d) for d in response]

    def create_dataset(self, dataset_name: str, description: str = "") -> Dataset:
        """Create a new dataset

        Args:
            dataset_name: name of the dataset. Must be unique within the namespace
            description: A description of the dataset

        Returns:
            A populated Dataset instance
        """
        data = {"name": dataset_name, "description": description}
        response = self._request("POST", f"/namespace/{self.name}/dataset/", json=data)
        return self._populate_dataset(response)

    def get_dataset(self, dataset_name: str) -> Dataset:
        """Get a Dataset instance by name

        Args:
            dataset_name: The name of the dataset

        Returns:
            A populated Dataset instance
        """
        response = self._request("GET", f"/namespace/{self.name}/dataset/{dataset_name}")
        return self._populate_dataset(response)

    def delete_dataset(self, dataset_name: str) -> None:
        """

        Args:
            dataset_name: The name of the dataset to delete

        Returns:
            None
        """
        self._request("DELETE", f"/namespace/{self.name}/dataset/{dataset_name}")





from typing import Optional, List, Dict

import hoss
from hoss.error import *
from hoss.auth import AuthService
from hoss.namespace import Namespace
from hoss.objectstore import ObjectStore
from hoss.api import CoreAPI


class CoreService(CoreAPI):
    """Class to interact a Hoss Server's core API. This is typically the entrypoint to the rest of the resources."""
    def __init__(self, server_url: str, auth_instance: Optional[AuthService] = None):
        """A class to interact with the Hoss core service

        Args:
            server_url: root URL to the server, including the protocol
            auth_instance: an optional AuthService instance that will be used instead of using the server_url to
             instantiate one
        """
        super().__init__(server_url, auth_instance)

    def __repr__(self):
        return f"<Hoss Server: {self.base_url}>"

    def _populate_object_store(self, api_response: dict) -> ObjectStore:
        """Helper function to parse an API response into an ObjectStore instance

        Args:
            api_response: The API response for an `object_store`.

        Returns:
            A populated ObjectStore instance
        """
        return ObjectStore(api_response.get('name'),
                           api_response.get('description'),
                           api_response.get('endpoint'),
                           api_response.get('type'),
                           self.base_url,
                           self.auth)

    def list_object_stores(self):
        """List Object Stores within the Server

        Returns:
            list of object store instances
        """
        response = self._request("GET", f"/object_store/")
        return [self._populate_object_store(n) for n in response]

    def get_object_store(self, name: str):
        """Get Object Stores by name

        Returns:
            list of object store instances
        """
        response = self._request("GET", f"/object_store/{name}")
        return self._populate_object_store(response)

    def _populate_namespace(self, api_response: dict) -> Namespace:
        """Helper function to parse an API response into an Namespace instance

        Args:
            api_response: The API response for a`namespace`.

        Returns:
            A populated Namespace instance
        """
        return Namespace(self.base_url, api_response.get('name'), api_response.get('bucket_name'),
                         api_response.get('description'),
                         self._populate_object_store(api_response.get('object_store')),
                         self.auth)

    def list_namespaces(self) -> List[Namespace]:
        """List Namespaces within the Server

        Returns:
            list of namespace instances
        """
        response = self._request("GET", f"/namespace/")
        return [self._populate_namespace(n) for n in response]

    def get_namespace(self, name: str) -> Namespace:
        """Get a namespace by name

        Returns:
            list of namespace instances
        """
        response = self._request("GET", f"/namespace/{name}")
        return self._populate_namespace(response)

    def create_namespace(self, name: str, description: str, object_store_name: str, bucket_name: str) -> Namespace:
        """Create a new namespace

        Args:
            name: name of the namespace
            description: description of the namespace
            object_store_name: name of the object store the namespace uses
            bucket_name: name of the bucket in the object store

        Returns:
            A populated Namespace instance
        """
        data = {"name": name, "description": description,
                "object_store_name": object_store_name, "bucket_name": bucket_name}
        response = self._request("POST", f"/namespace/", json=data)
        return self._populate_namespace(response)

    def delete_namespace(self, name: str) -> None:
        """Delete a namespace. Note, the namespace must be empty to delete (i.e. contains no datasets).

        Args:
            name: name of the namespace

        Returns:
            None
        """
        dataset_list = self._request("GET", f"/namespace/{name}/dataset/")
        if len(dataset_list) > 0:
            raise HossException(f"The namespace '{name}' contains {len(dataset_list)} dataset. "
                               f"Namespace must be empty to delete")
        self._request("DELETE", f"/namespace/{name}")
        return

    def search(self, metadata: Dict[str, str], namespace: Optional[str] = None, dataset: Optional[str] = None,
               modified_before: Optional[str] = None, modified_after: Optional[str] = None,
               limit: Optional[int] = 25, offset: Optional[int] = 0) -> List:
        """Search a server for objects, returning raw responses.

        This function will return raw search results, which will be a list of dictionaries with the format:

             'uri': 'hoss+https://hoss.myserver.com:default:example-ds/folder1/my-file3.bin',
             'file_path': 'folder1/my-file3.bin',
             'dataset': 'example-ds',
             'namespace': 'default',
             'last_modified_date': '2021-12-14T18:21:40.760Z',
             'size_bytes': 435,
             'metadata': [{'foo': 'bar'}]}

        Args:
            metadata: dictionary of key-value pairs that must match. empty dict will return all values.
            namespace: name of namespace to filter results
            dataset: name of dataset to filter results (namespace must be set
                along with dataset to be valid)
            modified_before: datetime string format `2006-01-02T15:04:05.000Z` to filter results
            modified_after: datetime string format `2006-01-02T15:04:05.000Z` to filter results
            limit: number of items to return per page
            offset: starting point in the index for returned items

        Returns:
            List of dictionaries containg the raw search response
        """
        query_params = []

        # add metadata key pairs
        metadata_params = [f"{k}:{metadata[k]}" for k in metadata]
        if metadata_params:
            query_params.append("metadata=" + ",".join(metadata_params).lower())

        # add namespace and dataset filters
        if namespace:
            query_params.append(f"namespace={namespace}")
            if dataset:
                query_params.append(f"dataset={dataset}")
        elif dataset:
            raise HossException("Must provide a namespace if searching within a specific dataset")
        
        # add time range filters
        if modified_before:
            query_params.append(f"modified_before={modified_before}")
        if modified_after:
            query_params.append(f"modified_after={modified_after}")

        # add pagination
        query_params.append(f"size={limit}")
        query_params.append(f"from={offset}")

        query_param_str = "&".join(query_params)

        # Remember all metadata headers are case insensitive, so we lowercase everything
        response = self._request("GET", f"/search?{query_param_str}")
        return response['results']

    def search_refs(self, metadata: Dict[str, str], namespace: Optional[str] = None, dataset: Optional[str] = None,
                    modified_before: Optional[str] = None, modified_after: Optional[str] = None,
                    limit: Optional[int] = 25, offset: Optional[int] = 0) -> List:
        """Search a server for objects, returning populated DatasetRef instances.

        Args:
            metadata: dictionary of key-value pairs that must match. empty dict will return all values.
            namespace: name of namespace to filter results
            dataset: name of dataset to filter results (namespace must be set
                along with dataset to be valid)
            modified_before: datetime string format `2006-01-02T15:04:05.000Z` to filter results
            modified_after: datetime string format `2006-01-02T15:04:05.000Z` to filter results
            limit: number of items to return per page
            offset: starting point in the index for returned items

        Returns:
            A list of populated DatasetRefs
        """

        results = self.search(metadata, namespace, dataset, modified_before, modified_after, limit, offset)
        return [hoss.resolve(result["uri"], self.auth) for result in results]



from typing import Optional

import boto3
from boto3.s3.transfer import TransferConfig, MB
import botocore.exceptions
import datetime

from hoss.api import CoreAPI
from hoss.auth import AuthService


class ObjectStore(CoreAPI):
    """A class representing an Object Store that provides low-level functionality. Typically you should be interfacing
     with DatasetRef instances."""
    def __init__(self, object_store_name: str, description: str, endpoint: str, object_store_type: str,
                 server_url: str, auth_instance: Optional[AuthService] = None):
        """A class to interact with the Hoss core service
        
        Args:
            object_store_name: name of the object store
            description: description of the store
            endpoint: object store endpoint (e.g. http://localhost, https://s3.amazonaws.com)
            object_store_type: type of object store (e.g. "s3", "minio")
        """
        super().__init__(server_url, auth_instance)

        self.name = object_store_name
        self.description = description
        self.endpoint = endpoint
        self.object_store_type = object_store_type

        # S3 client
        self._client = dict()
        self._transfer_config = TransferConfig()
        self._sts_credential_expire = None
        self._sts_credential_expire_in_seconds = self.auth.jwt_exp_seconds / 2

    def __repr__(self):
        return f"<Object Store: {self.name} - {self.description} ({self.endpoint})>"

    def _get_sts_credentials(self, namespace: str) -> dict:
        return self._request("GET", f"/namespace/{namespace}/sts")

    def _get_client(self, namespace: str):
        if self._client.get(namespace) is not None:
            if self._sts_credential_expire > datetime.datetime.utcnow():
                # Credentials are still good. Return existing client.
                return self._client.get(namespace)

        # Fetch temporary credentials and create an S3 client
        creds = self._get_sts_credentials(namespace)
        self._sts_credential_expire = datetime.datetime.utcnow() + \
                                      datetime.timedelta(seconds=self._sts_credential_expire_in_seconds)

        client = boto3.client(
            "s3",
            endpoint_url=creds["endpoint"],
            aws_access_key_id=creds["access_key_id"],
            aws_secret_access_key=creds["secret_access_key"],
            aws_session_token=creds["session_token"],
            region_name=creds["region"]
        )
        self._client[namespace] = client
        return client

    def object_exists(self, namespace: str, bucket: str, key) -> bool:
        """Check if an object exists in the object store

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key

        Returns:
            True if the object exists, False if it does not
        """
        # TODO: Need to handle grabbing metadata here better
        client = self._get_client(namespace)

        try:
            client.head_object(Bucket=bucket, Key=key)
            return True
        except botocore.exceptions.ClientError as ce:
            code = ce.response["Error"]["Code"]
            if code == '404':
                return False
            raise

    def head_object(self, namespace: str, bucket: str, key) -> dict:
        """Method to perform a HEAD operation on an object

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key

        Returns:
            Dict with the fields `last_modified`, `size_bytes`, `etag`, and `metadata`
        """
        client = self._get_client(namespace)
        response = client.head_object(Bucket=bucket, Key=key)
        return {"last_modified": response.get('LastModified'),
                "size_bytes": response.get('ContentLength'),
                "etag": response.get('ETag'),
                "metadata": response.get('Metadata')}

    @staticmethod
    def _remove_prefix(text, prefix):
        return text[text.startswith(prefix) and len(prefix):]

    def list_objects(self, namespace: str, bucket: str, prefix, recursive=False) -> dict:
        """Method to list objects in a bucket, optionally with a prefix and recursively

        While iterating, we use the `/` char as a delimiter to represent "folders" within the store

        Args:
            namespace: Name of the namespace
            bucket: Name of the bucket
            prefix: key prefix to start listing from
            recursive: If True, continue to list objects "down" the tree, instead of staying within the specified
                       "directory" level

        Yields:
            a dict with the fields `key`, `name`, `last_modified`, `size_bytes`, and `etag`

        """
        client = self._get_client(namespace)

        kwargs = {
            "Bucket": bucket,
            "Delimiter": "/",
        }

        if recursive:
            del kwargs["Delimiter"]

        if not prefix.endswith('/'):
            prefix += '/'
        kwargs["Prefix"] = prefix

        paginator = client.get_paginator("list_objects_v2")
        for resp in paginator.paginate(**kwargs):
            for d in resp.get("CommonPrefixes", []):
                yield {"key": d["Prefix"],
                       "name": self._remove_prefix(d["Prefix"], prefix),
                       "etag": d.get("ETag"),
                       "last_modified": d.get("LastModified"),
                       "size_bytes": d.get("Size")}

            for f in resp.get("Contents", []):
                name = self._remove_prefix(f["Key"], prefix)
                if name == ".dataset.yaml":
                    # Do not show the user the internal Hoss '.dataset.yaml' file.
                    continue

                yield {"key": f["Key"],
                       "name": name,
                       "etag": f.get("ETag"),
                       "last_modified": f.get("LastModified"),
                       "size_bytes": f.get("Size")}

    def get_object(self, namespace: str, bucket: str, key):
        """Method to get an object's contents from the object store

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key

        Returns:
            contents of the object
        """
        # TODO: Need to handle grabbing metadata here better
        client = self._get_client(namespace)
        return client.get_object(Bucket=bucket, Key=key)["Body"].read()

    def put_object(self, namespace: str, bucket: str, key: str,
                   data: bytes, metadata: Optional[dict] = None) -> None:
        """Method to write an object to the object store

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key
            data: The data to write
            metadata: A dictionary of key-value pairs to set as object metadata

        Returns:
            None
        """
        client = self._get_client(namespace)
        client.put_object(Bucket=bucket, Key=key, Body=data,
                          Metadata=metadata if metadata is not None else dict())

    def delete_object(self, namespace: str, bucket: str, key):
        """Method to delete an object from the object store

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key

        Returns:
            None
        """
        client = self._get_client(namespace)
        client.delete_object(Bucket=bucket, Key=key)

    def copy_object(self, namespace: str, bucket: str, key: str, target_key: str) -> None:
        """Method to perform a copy an object from one key to another in the object store

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key
            target_key: Destination object key

        Returns:
            None
        """
        client = self._get_client(namespace)
        source = {"Bucket": bucket, "Key": key}
        client.copy_object(Bucket=bucket, Key=target_key, CopySource=source)

    def download_file(self, namespace: str, bucket: str, key: str, filename: str):
        """Method to download an object to a file locally

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Source object key
            filename: Absolute path to a file where the object's contents will be written

        Returns:

        """
        client = self._get_client(namespace)
        return client.download_file(bucket, key, filename, Config=self._transfer_config)

    def download_fileobj(self, namespace: str, bucket: str, key, fh):
        """Download an object to a file-like object.

        The file-like object must be in binary mode.
        This is a managed transfer which will perform a multipart download in multiple threads if necessary.

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key
            fh: A file handle to a file-like object in binary mode

        Returns:

        """
        client = self._get_client(namespace)
        return client.download_fileobj(bucket, key, fh, Config=self._transfer_config)

    def upload_file(self, namespace: str, bucket: str, key, filename, metadata: Optional[dict] = None):
        """Upload a file to an object

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key
            filename: Absolute path to the file to upload
            metadata: A dictionary of key-value pairs to set as object metadata

        Returns:

        """
        client = self._get_client(namespace)
        return client.upload_file(filename, bucket, key, Config=self._transfer_config,
                                  ExtraArgs={"Metadata": metadata} if metadata is not None else None)

    def upload_fileobj(self, namespace: str, bucket: str, key, fh, metadata: Optional[dict] = None):
        """Upload a file-like object to an object

        Args:
            namespace: Namespace name
            bucket: Bucket name
            key: Object key
            fh: File handle to a file-like object in binary mode
            metadata: A dictionary of key-value pairs to set as object metadata

        Returns:

        """
        client = self._get_client(namespace)
        return client.upload_fileobj(fh, bucket, key, Config=self._transfer_config,
                                     ExtraArgs={"Metadata": metadata} if metadata is not None else None)

    def set_transfer_config(self, multipart_threshold: int, max_concurrency: int,
                            multipart_chunksize: int = 8*MB) -> None:
        """Function to set the transfer configuration used in managed transfers.

        Args:
            multipart_threshold: The transfer size threshold for which multipart uploads, downloads,
                                 and copies will automatically be triggered.
            max_concurrency: The maximum number of threads that will be making requests to perform a transfer.
            multipart_chunksize: The partition size of each part for a multipart transfer. (default 8MB)

        Returns:

        """
        self._transfer_config = TransferConfig(
            multipart_threshold=multipart_threshold,
            max_concurrency=max_concurrency,
            multipart_chunksize=multipart_chunksize
        )

    def get_multipart_chunk_size(self) -> int:
        """Get the current multi-part chunk size

        Returns:
            The multi-part chunk size in bytes
        """
        return self._transfer_config.multipart_chunksize

    def get_multipart_threshold(self) -> int:
        """Get the current multi-part threshold

        Returns:
            The multi-part threshold in bytes
        """
        return self._transfer_config.multipart_threshold



from setuptools import setup, find_packages
import codecs
import os.path


# to publish
# python setup.py sdist bdist_wheel
# twine upload --skip-existing dist/*


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='hoss-client',
    version=get_version("hoss/version.py"),
    description='A client library and CLI for interacting with a Hoss server.',
    long_description_content_type='text/markdown',
    long_description=read('README.md'),
    author='Gigantum, Inc.',
    author_email='support@gigantum.com',

    install_requires=[
        "requests",
        "boto3",
        "PyJWT",
        "click",
        "rich",
        "humanize",
        "packaging"
    ],
    entry_points={
        'console_scripts': ['hoss=hoss.cli:cli'],
    },
    packages=find_packages(),
    python_requires='>=3.6',
    include_package_data=True,

    url='https://github.com/gigantum/hoss-client',
    license='MIT',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.6',
    ],
    keywords=[
        'data-science',
        'science',
        'gigantum',
        'open-science',
        'object-store',
        's3',
        'minio',
    ]
)



from typing import List, Optional, Dict
import hoss

import requests
import re

import os
import time
import datetime
from pathlib import Path
from enum import Enum

from rich import box
from rich.panel import Panel
from rich.progress import Progress, SpinnerColumn, BarColumn, TextColumn
from rich.style import Style
from rich.text import Text
from rich.live import Live
from rich.table import Table
from rich.align import Align
from rich.layout import Layout

import humanize
from boto3.s3.transfer import MB

from hoss.console import console


class TaskStatus(Enum):
    """Enum for setting the state of tasks"""
    WAITING = 0
    RUNNING = 1
    DONE = 2
    ERROR = -1


def _get_task_status(tasks: List[str], status: List[TaskStatus]) -> Align:
    """Take a list of task statuses and render a table

    Args:
        tasks: list of tasks
        status: list of TaskStatus values

    Returns:
        the rendered Table
    """
    table = Table(box=box.SIMPLE_HEAVY, show_footer=False)
    table_centered = Align.center(table)

    table.add_column("Status", no_wrap=True)
    table.add_column("Tasks", no_wrap=True)

    for t, s in zip(tasks, status):
        if s == TaskStatus.WAITING:
            table.add_row(":hourglass:", t)
        elif s == TaskStatus.RUNNING:
            table.add_row(":face_with_monocle:", t)
        elif s == TaskStatus.DONE:
            table.add_row(":white_check_mark:", t)
        elif s == TaskStatus.ERROR:
            table.add_row(":x:", t)
        else:
            table.add_row(":question:", t)

    return table_centered


def upload_directory(dataset_name: str, directory: str, namespace: str, endpoint: str,
                     skip: str, max_concurrency: int = 10, multipart_threshold: int = 32,
                     multipart_chunk_size: int = 32, metadata: Optional[Dict[str, str]] = None) -> None:
    """Function to upload a directory with a CLI interface for status

    This method will sort files by size, starting with the smallest first, and then upload them
    one at a time.

    Args:
        dataset_name: Name of the dataset to upload into
        directory: The containing the files to upload. The directory name will be included.
        namespace: The namespace containing the dataset
        endpoint: The URL for the Hoss server you are uploading to
        skip: A regex string of files that should be skipped
        max_concurrency: max concurrency
        multipart_threshold: threshold in megabytes to activate multipart uploads
        multipart_chunk_size: size in megabytes to use multipart uploads
        metadata: Optional dict of key-value pairs to write to all files as metadata

    Returns:
        None
    """
    log_msgs = list()

    def _log(msg: str, layout_obj: Layout, update_last=False) -> None:
        """Helper method to update and build the log window contents"""
        if update_last:
            del log_msgs[0]

        log_msgs.insert(0, f"{datetime.datetime.now().strftime('%H:%M:%S.%f %p')} - {msg}")
        log_panel = Panel(Text("\n".join(log_msgs)), title="[b]Status", border_style="#957299", padding=(1, 2),
                          expand=True)
        layout_obj["status"].update(log_panel)

    # Initialize the layout
    console.clear()
    layout = Layout(name="root")
    layout.split(
        Layout(name="title", size=3),
        Layout(name="header", size=10),
        Layout(name="main", ratio=1),
    )
    layout["header"].split_row(
        Layout(name="summary"),
        Layout(name="tasks", ratio=1, minimum_size=60),
    )
    layout["main"].split_column(
        Layout(name="progress", size=10),
        Layout(name="status"),
    )

    # Set title element
    title = Table.grid(expand=True)
    title.add_column(justify="center", ratio=1)
    title.add_row("[b]Hoss Upload Tool[/b]")
    layout["title"].update(Panel(title, style="white on #957299"))

    # Set summary element
    summary_grid = Table(box=box.SIMPLE_HEAVY, show_footer=False)
    summary_grid_centered = Align.center(summary_grid)
    summary_grid.add_column(justify="right")
    summary_grid.add_column("Settings", justify="left")
    summary_grid.add_row("[b]Server:[/b]", endpoint)
    summary_grid.add_row("[b]Namespace:[/b]", namespace)
    summary_grid.add_row("[b]Dataset:[/b]", dataset_name)
    summary_grid.add_row("[b]Directory:[/b]", directory)
    layout["summary"].update(summary_grid_centered)

    tasks = ["Check for credentials",
             "Check server connectivity",
             "Check dataset exists",
             "Process directory",
             "Upload data",
             ]

    layout["tasks"].update(_get_task_status(tasks,
                                            [TaskStatus.WAITING,
                                             TaskStatus.WAITING,
                                             TaskStatus.WAITING,
                                             TaskStatus.WAITING,
                                             TaskStatus.WAITING]))

    # Initialize Progress section
    job_progress = Progress(
        "{task.description}",
        SpinnerColumn(),
        BarColumn(pulse_style=Style(color="#957299"), complete_style=Style(color="#957299")),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%", style=Style(color="#957299")),
        expand=True
    )
    prepare_job_id = job_progress.add_task("[white]Preparing Upload", total=7)
    upload_job_id = job_progress.add_task("[white]Uploading Data", start=False)
    layout['progress'].update(
        Align.center(Panel(job_progress, title="[b]Progress", border_style="#957299", padding=(1, 2), expand=False),
                     vertical="middle"))

    _log("Starting...", layout)

    with Live(layout, console=console, refresh_per_second=30):
        # Check for PAT
        if not os.environ.get("HOSS_PAT"):
            # No credentials set
            _log("Personal Access Token not found!", layout)
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.ERROR,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING]))
            layout['progress'].update(
                Align.center(Panel("Failed to get personal access token from environment variable.\n\n"
                                   "Make sure you have set the `HOS_PAT` environment variable and try again."),
                             style="white on red", height=6))
            return
        else:
            _log("Personal Access Token found.", layout)
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.RUNNING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING]))
            job_progress.advance(prepare_job_id)
            time.sleep(1)

        try:
            _log("Connecting to server and exchanging credentials...", layout)
            server = hoss.connect(endpoint)
            ns = server.get_namespace(namespace)
            _log("Server reachable and credentials ready.", layout)
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.DONE,
                                                     TaskStatus.RUNNING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING]))
            job_progress.advance(prepare_job_id)
        except requests.exceptions.ConnectionError:
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.ERROR,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING]))
            _log(f"Server unreachable. Verify your network connection or endpoint setting.", layout)
            layout['progress'].update(
                Align.center(Panel("Failed to reach server. \n\nCheck your endpoint and network connectivity.",
                                   style="white on red", height=6)))
            return
        except hoss.HossException as err:
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.ERROR,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING,
                                                     TaskStatus.WAITING]))
            layout['progress'].update(Align.center(Panel(
                f"Failed to exchange credentials. \n\nVerify PAT with your Hoss server and try again. Error: \n\n {err}",
                style="white on red", height=10)))
            _log(f"Failed to exchange credentials.", layout)
            return

        try:
            _log(f"Checking if dataset exists in the specified server and namespace.", layout)
            ds = ns.get_dataset(dataset_name)
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.DONE,
                                                     TaskStatus.DONE,
                                                     TaskStatus.RUNNING,
                                                     TaskStatus.WAITING]))
            _log(f"Dataset ready.", layout)
            job_progress.advance(prepare_job_id)
        except hoss.NotFoundException:
            # Dataset does not exist
            layout['progress'].update(Align.center(
                Panel(f"Specified dataset does not exist. \n\nVerify input or create dataset and try again.",
                      style="white on red", height=6)))
            layout["tasks"].update(_get_task_status(tasks,
                                                    [TaskStatus.DONE,
                                                     TaskStatus.DONE,
                                                     TaskStatus.DONE,
                                                     TaskStatus.ERROR,
                                                     TaskStatus.WAITING]))
            _log(f"Dataset not found. Please create it and try again.", layout)
            return

        # set transfer settings
        ds.namespace.object_store.set_transfer_config(multipart_threshold * MB,
                                                      max_concurrency,
                                                      multipart_chunk_size * MB)

        job_progress.advance(prepare_job_id)

        # list files
        _log(f"Scanning directory.", layout)
        files = filter(os.path.isfile, Path(directory).rglob('*'))
        job_progress.advance(prepare_job_id)

        # Sort files
        _log(f"Sorting files.", layout)
        files_sorted = sorted(files, key=lambda x: os.stat(x).st_size)
        job_progress.advance(prepare_job_id)

        # Find files to skip if you need to
        if skip != "":
            _log(f"Processing skipped files.", layout)
            skip_check = re.compile(r'{}'.format(skip))
            skip_files = [f for f in files_sorted if skip_check.match(f.name)]
            for f in skip_files:
                files_sorted.remove(f)
                if not skip:
                    files_sorted.append(f)

        job_progress.advance(prepare_job_id)
        layout["tasks"].update(_get_task_status(tasks,
                                                [TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.WAITING]))

        job_progress.update(upload_job_id, total=len(files_sorted), description="[white]Uploading Data")
        job_progress.start_task(upload_job_id)
        _log(f"Uploading {len(files_sorted)} files.", layout)

        # upload files
        errors = False
        session_folder = os.path.basename(directory)
        for cnt, f in enumerate(files_sorted):
            if f.name == ".DS_Store":
                _log(f"Skipping {f.relative_to(directory)}.", layout)
                job_progress.advance(upload_job_id)
                continue

            try:
                o = ds / session_folder / f.relative_to(directory).as_posix()
                if not o.exists():
                    _log(f"[{cnt+1}/{len(files_sorted)}] Uploading {f.relative_to(directory)}..."
                         f" ({humanize.naturalsize(f.stat().st_size)}).", layout)
                    o.write_from(f.as_posix(), metadata=metadata)
                    _log(f"[{cnt + 1}/{len(files_sorted)}] Uploading {f.relative_to(directory)}...done!"
                         f" ({humanize.naturalsize(f.stat().st_size)}).", layout, update_last=True)
                else:
                    _log(f"[{cnt+1}/{len(files_sorted)}] Skipping {f.relative_to(directory)}. File already exists.",
                         layout)
            except Exception as err:
                _log(f"ERROR: {f.relative_to(directory)} upload failed!", layout)
                _log(str(err), layout)
                errors = True

            job_progress.advance(upload_job_id)

        if not errors:
            _log(f"Upload complete!.", layout)
        else:
            _log(f"Upload completed with errors! Check for missing files.", layout)
        layout["tasks"].update(_get_task_status(tasks,
                                                [TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.DONE,
                                                 TaskStatus.DONE]))

        for cnt in range(15):
            _log(f"Exiting in {15-cnt} seconds...", layout, update_last=True)
            time.sleep(1)

    console.clear()



from typing import TYPE_CHECKING

import datetime

from hoss.api import CoreAPI
from hoss.ref import DatasetRef
import hoss.error

if TYPE_CHECKING:
    from hoss.namespace import Namespace

PERM_NONE = None
PERM_READ = "r"
PERM_READ_WRITE = "rw"


class Dataset(CoreAPI, DatasetRef):
    def __init__(self, namespace: 'Namespace', dataset_name: str, description: str,
                 created_on: datetime.datetime,
                 root_directory: str, owner: str, bucket: str, permissions: dict) -> None:
        self.namespace = namespace
        self.dataset_name = dataset_name
        self.description = description
        self.created_on = created_on
        self.root_directory = root_directory
        self.owner = owner
        self.bucket = bucket
        self.permissions = permissions

        CoreAPI.__init__(self, server_url=self.namespace.base_url, auth_instance=namespace.auth)
        DatasetRef.__init__(self, namespace=self.namespace, parent=None, dataset_name=dataset_name,
                            key=root_directory, name='', etag=None, last_modified=None, size_bytes=None)

    def __repr__(self) -> str:
        return f"<Dataset: {self.namespace} - {self.dataset_name}>"

    def display(self):
        """Helper to print some useful info about the dataset"""
        print("--Dataset------")
        print(f"Name: {self.dataset_name}")
        print(f"Description: {self.description}")
        print(f"Namespace: {self.namespace}")
        print(f"Created On: {self.created_on}")
        print(f"Root Directory: {self.root_directory}")
        print(f"Owner: {self.owner}")
        print("Permissions:")
        for u, p in self.permissions.items():
            print(f"\t{u}: {p}")
        print("---------------")
        return self

    def is_sync_enabled(self) -> bool:
        """Check if sync is enabled for the dataset"""
        response = self._request("GET", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/sync")
        return response

    def enable_sync(self, sync_type: str) -> None:
        """Method to enable syncing on a dataset. The namespace must also already have syncing enabled.

        If the namespace is configured with duplex syncing, you can set the dataset to either simplex or duplex
        If the namespace is configured with only simplex syncing, you can only set the dataset to simplex

        Note - currently syncing only will sync objects mutated after syncing is enabled.

        Args:
            sync_type: type of sync config (`simplex` or `duplex`)

        Returns:
            None
        """
        if sync_type not in ["simplex", "duplex"]:
            raise hoss.error.HossException(f"'sync_type' must be either 'simplex' or 'duplex'")

        data = {"sync_type": sync_type}
        self._request("PUT", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/sync", json=data)

    def disable_sync(self) -> None:
        """Disable dataset syncing

        Returns:
            None
        """
        self._request("DELETE", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/sync")

    # Dataset admin methods
    def set_user_permission(self, username, permission):
        """Set permissions on the dataset for a user.

        This will attach the user's "personal" group under the hood.

        Args:
            username: username for the user you wish to set
            permission: permission level ('r', 'rw', or None)

        Returns:

        """
        if permission == PERM_NONE:
            self._request("DELETE", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/user/{username}")
            del self.permissions[username + "-hoss-default-group"]
        else:
            self._request("PUT", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/user/{username}/access/{permission}")
            self.permissions[username + "-hoss-default-group"] = permission
        return self
    
    def set_group_permission(self, group_name, permission):
        """Set permissions on the dataset for a groupd.

        This will attach the user's "personal" group under the hood.

        Args:
            group_name: group you wish to set
            permission: permission level ('r', 'rw', or None)

        Returns:

        """
        if permission == PERM_NONE:
            self._request("DELETE", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/group/{group_name}")
            del self.permissions[group_name]
        else:
            self._request("PUT", f"/namespace/{self.namespace.name}/dataset/{self.dataset_name}/group/{group_name}/access/{permission}")
            self.permissions[group_name] = permission
        return self



import click

from hoss.commands.upload import upload
from hoss.commands.version import version

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(help="A Command Line Interface to interact with a Hoss server.",
             context_settings=CONTEXT_SETTINGS)
def cli():
    pass


# Add commands from package
cli.add_command(upload)
cli.add_command(version)


if __name__ == '__main__':
    cli()
